@@ -28,15 +28,15 @@ use fields qw( code text header lines body as_string );
 ###########################################################################
 # Constructor
 # Creates new object. If there was only one argument it will interprete
-# it as a string representation (see new_from_string), otherwise it will
+# it as a string representation (see new_from_string), otherwise it will 
 # assume a hash/array representation (see new_from_parts)
 # Args: see new_from_string|new_from_parts
 # Returns: $self
 ###########################################################################
 sub new {
 	my $class = shift;
-	return @_>1
-		? $class->new_from_parts(@_)
+	return @_>1 
+		? $class->new_from_parts(@_) 
 		: $class->new_from_string(@_);
 }
 
@@ -47,7 +47,7 @@ sub new {
 #   $text:   Response text or request URI
 #   $header: Header representation as array or hash
 #            either [ [key1 => val2],[key2 => val2],... ] where the same
-#            key can occure multiple times
+#            key can occure multiple times 
 #            or { key1 => val1, key2 => val2 } where val can be either
 #            a scalar or an array-ref (if the same key has multiple values)
 #   $body:   Body as string
@@ -74,18 +74,22 @@ sub new_from_parts {
 		$header = \@hnew;
 	}
 
+	my $self = fields::new($class);
+	my $rebless;
 	if ( $code =~m{^\d} ) {
 		# Response
-		$class = 'Net::SIP::Response' if $class eq 'Net::SIP::Packet';
+		$self->{code} = $code;
+		$self->{text} = defined($text) ? $text:'';
+		$rebless = 'Net::SIP::Response';
 	} else {
 		# Request
-		$code = uc($code);                             # uppercase method
-		$class = 'Net::SIP::Request' if $class eq 'Net::SIP::Packet';
+		$self->{code} = uc($code);                             # uppercase method
+		$self->{text} = defined($text) ? $text:'';
+		$rebless = 'Net::SIP::Request';
 	}
 
-	my $self = fields::new($class);
-	$self->{code} = $code;
-	$self->{text} = defined($text) ? $text:'';
+	# rebless to Net::SIP::{Request,Response}
+	bless $self,$rebless if $class eq 'Net::SIP::Packet';
 
 	# $self->{header} is list of Net::SIP::HeaderPair which cares about normalized
 	# keys while maintaining the original key, so that one can restore header
@@ -111,7 +115,7 @@ sub new_from_parts {
 
 	if ( ref($body)) {
 		if ( !$self->get_header( 'content-type' )) {
-			my $sub = UNIVERSAL::can( $body, 'content_type' );
+			my $sub = UNIVERSAL::can( $body, 'content-type' );
 			$self->set_header( 'content-type' => $sub->($body) ) if $sub;
 		}
 		$body = $body->as_string;
@@ -131,14 +135,14 @@ sub new_from_parts {
 ###########################################################################
 sub new_from_string {
 	my ($class,$string) = @_;
-	my $data = _string2parts( $string );
-	if ( $class eq 'Net::SIP::Packet' ) {
-		$class = $data->{code} =~m{^\d} 
-			? 'Net::SIP::Response'
-			:'Net::SIP::Request';
-	}
 	my $self = fields::new($class);
-	%$self = %$data;
+	$self->{as_string} = $string;
+	if ( $class eq 'Net::SIP::Packet' ) {
+		# rebless
+		# as a side effect is_request will parse string so that code,header etc
+		# will be set
+		bless $self,( $self->is_request ? 'Net::SIP::Request':'Net::SIP::Response' );
+	}
 	return $self;
 }
 
@@ -158,8 +162,8 @@ sub is_request {
 # Args: $self
 # Returns: 1 if it's a response
 ###########################################################################
-sub is_response {
-	return ! shift->is_request()
+sub is_response { 
+	return ! shift->is_request() 
 }
 
 
@@ -202,7 +206,7 @@ sub callid { scalar( shift->get_header('call-id')) }
 # Returns: @val|\%header
 #   @val: if key given returns all values for this key
 #      croak()s if in scalar context and I've more then one value for the key
-#   \%header: if no key given returns hash with
+#   \%header: if no key given returns hash with 
 #      { key1 => \@val1, key2 => \@val2,.. }
 ###########################################################################
 sub get_header {
@@ -215,14 +219,7 @@ sub get_header {
 			push @v,$h->{value} if $h->{key} eq $key;
 		}
 		return @v if wantarray;
-		if (@v>1) {
-			# looks like we have multiple headers but expect only
-			# one. Because we've seen bad client which issue multiple
-			# content-length header we try if all in @v are the same
-			my %v = map { $_ => 1 } @v;
-			return $v[0] if keys(%v) == 1; # ok, only one
-			croak( "multiple values for $key" );
-		}
+		croak( "multiple values for $key" ) if @v>1;
 		return $v[0];
 	} else {
 		my %result;
@@ -242,7 +239,7 @@ sub get_header {
 # Returns: @val|\%header
 #   @val: if key given returns all values (Net::SIP::HeaderVal) for this key
 #      croak()s if in scalar context and I've more then one value for the key
-#   \%header: if no key given returns hash with
+#   \%header: if no key given returns hash with 
 #      { key1 => \@val1, key2 => \@val2,.. } where val are Net::SIP::HeaderVal
 ###########################################################################
 sub get_header_hashval {
@@ -261,7 +258,7 @@ sub get_header_hashval {
 	} else {
 		my %result;
 		foreach my $h (@$hdr) {
-			push @{ $result{$h->{key}} },
+			push @{ $result{$h->{key}} }, 
 				Net::SIP::HeaderVal->new( $h );
 		}
 		return \%result;
@@ -342,7 +339,7 @@ sub set_body {
 	my ($self,$body) = @_;
 	if ( ref($body)) {
 		if ( !$self->get_header( 'content-type' )) {
-			my $sub = UNIVERSAL::can( $body, 'content_type' );
+			my $sub = UNIVERSAL::can( $body, 'content-type' );
 			$self->set_header( 'content-type' => $sub->($body) ) if $sub;
 		}
 		$body = $body->as_string;
@@ -398,7 +395,7 @@ sub as_string {
 	}
 
 	# return immediatly if request is up to date
-	return $self->{as_string} if $self->{as_string};
+	return $self->{as_string} if $self->{as_string}; 
 
 	my $header = $self->{header};
 
@@ -410,7 +407,7 @@ sub as_string {
 			my ($line,$count) = @{ $lines->[$i] || next };
 
 			# check if $count entries for line-index $i in headers
-			my @hi = grep {
+			my @hi = grep { 
 				my $line = $header->[$_]{line};
 				( defined($line) && $line == $i ) ? 1:0;
 			} (0..$#$header);
@@ -437,7 +434,7 @@ sub as_string {
 
 		my $v = $header->[$hi]{value};
 		$v =~s{\r?\n\s*}{\r\n }g; # \r?\n\s* -> \r\n + space for continuation lines
-		$result[$hi] = ucfirst($header->[$hi]{key}).": ".$v;
+		$result[$hi] = $header->[$hi]{key}.": ".$v;
 	}
 
 	# (re)build packet
@@ -455,179 +452,102 @@ sub as_string {
 	return ( $self->{as_string} = $hdr_string."\r\n".$body );
 }
 
-###########################################################################
-# packet dump in long or short form, used mainly for debuging
-# Args: ($self,?$level)
-#  $level: level of details: undef|0 -> one line, else -> as_string
-# Returns: $dump_as_string
-###########################################################################
-sub dump {
-	my Net::SIP::Packet $self = shift;
-	my $level = shift;
-	if ( !$level ) {
-		my ($code,$text,$header,$body) = $self->as_parts;
-		if ( $self->is_request ) {
-			return "REQ  $code $text ".( $body ? 'with body' :'' );
-		} else {
-			return "RESP $code '$text' ".( $body ? 'with body' :'' );
-		}
-	} else {
-		return $self->as_string
-	}
-}
-
-
 ###########################################################################
 # Return parts
 # Args: ($self)
 # Returns: ($code,$text,$header,$body)
 #   $code:   Response code or request method
 #   $text:   Response text or request URI
-#   $header: Header representation as array
+#   $header: Header representation as array 
 #            [ [key1 => val2],[key2 => val2],... ] where the same
-#            key can occure multiple times
+#            key can occure multiple times 
 #   $body:   Body as string
 # Comment:
 # Output from this method is directly usable as input to new_from_parts
 ###########################################################################
 sub as_parts {
 	my $self = shift;
-
-	# if parts are up to date return immediatly#
-	if ( ! $self->{code} ) {
-		my $data = _string2parts( $self->{as_string} );
-		%$self = ( %$self,%$data );
-	}
+	
+	# if parts are up to date return immediatly
 	return @{$self}{qw(code text header body)} if $self->{code};
-}
 
-{
-	my $word_rx = qr{[\w\-\.!%\*+`'~()<>:"/?{}\x1c\x1b\x1d]+};
-	my $callid_rx = qr{^$word_rx(?:\@$word_rx)?$};
-	my %key2parser = (
+	# otherwise parse request
+	my ($header,$body) = split( m{\r?\n\r?\n}, $self->{as_string},2 );
+	my @header = split( m{\r?\n}, $header );
 
-		# FIXME: More of these should be more strict to filter out invalid values
-		# for now they are only given here to distinguish them from the keys, which
-		# can be given multiple times either on different lines or on the same delimited
-		# by comma
-
-		'www-authenticate' => \&_hdrkey_parse_keep,
-		'authorization' => \&_hdrkey_parse_keep,
-		'proxy-authenticate' => \&_hdrkey_parse_keep,
-		'proxy-authorization' => \&_hdrkey_parse_keep,
-		'date' => \&_hdrkey_parse_keep,
-		'content-disposition' => \&_hdrkey_parse_keep,
-		'content-type' => \&_hdrkey_parse_keep,
-		'mime-version' => \&_hdrkey_parse_keep,
-		'organization' => \&_hdrkey_parse_keep,
-		'priority' => \&_hdrkey_parse_keep,
-		'reply-to' => \&_hdrkey_parse_keep,
-		'retry-after' => \&_hdrkey_parse_keep,
-		'server' => \&_hdrkey_parse_keep,
-		'to' => \&_hdrkey_parse_keep,
-		'user-agent' => \&_hdrkey_parse_keep,
-
-		'content-length' => \&_hdrkey_parse_num,
-		'expires' => \&_hdrkey_parse_num,
-		'max-forwards' => \&_hdrkey_parse_num,
-		'min-expires' => \&_hdrkey_parse_num,
-
-		'call-id' => sub { 
-			$_[0] =~ $callid_rx or die "invalid callid, should be 'word [@ word]'";
-			return $_[0];
-		},
-		'cseq' => sub { 
-			$_[0] =~ m{^\d+\s+\w+\s*$} or die "invalid cseq, should be 'number method'";
-			return $_[0];
-		},
-	);
-
-	sub _hdrkey_parse_keep { return $_[0] };
-	sub _hdrkey_parse_num {
-		my ($v,$k) = @_;
-		$v =~m{^(\d+)\s*$} || die "invalid $k, should be number";
-		return $1;
-	};
-
-	sub _hdrkey_parse_comma_seperated {
-		my ($v,$k) = @_;
-		my @v = ( '' );
-		my $quoted = 0;
-		# split on komma (but not if quoted)
-		while (1) {
-			if ( $v =~m{\G(.*?)([\\",])}gc ) {
-				if ( $2 eq "\\" ) {
-					$v[-1].=$1.$2.substr( $v,pos($v),1 );
-					pos($v)++;
-				} elsif ( $2 eq '"' ) {
-					$v[-1].=$1.$2;
-					$quoted = !$quoted;
-				} elsif ( $2 eq ',' ) {
-					# next item if not quoted
-					( $v[-1].=$1 ) =~s{\s+$}{}; # strip trailing space
-					push @v,'' if !$quoted;
-					$v =~m{\G\s+}gc; # skip space after ','
-				}
-			} else {
-				# add rest to last from @v
-				$v[-1].= substr($v,pos($v)||0 );
-				last;
-			}
-		}
-		return @v;
+	if ( $header[0] =~m{^SIP/2.0\s+(\d+)\s+(\S.*?)\s*$} ) {
+		# Response, e.g. SIP/2.0 407 Authorization required
+		$self->{code} = $1;
+		$self->{text} = $2;
+	} elsif ( $header[0] =~m{^(\w+)\s+(\S.*?)\s+SIP/2\.0\s*$} ) {
+		# Request, e.g. INVITE <sip:bla@fasel> SIP/2.0
+		$self->{code} = $1;
+		$self->{text} = $2;
+	} else {
+		die "bad request: starts with '$header[0]'";
 	}
+	shift(@header);
 
-	sub _string2parts {
-		my $string = shift;
-		my %result = ( as_string => $string );
+	$self->{body} = $body;
+	
+	my @hdr;
+	my @lines;
+	while (@header) {
+		my ($k,$v) = $header[0] =~m{^([^\s:]+)\s*:\s*(.*)}
+			or die "bad header line $header[0]";
+		my $line = shift(@header);
+		while ( @header && $header[0] =~m{^\s+(.*)} ) {
+			# continuation line
+			$v .= "\n$1";
+			$line .= shift(@header);
+		}
+		my $nk = _normalize_hdrkey($k);
 
-		# otherwise parse request
-		my ($header,$body) = split( m{\r?\n\r?\n}, $string,2 );
-		my @header = split( m{\r?\n}, $header );
-
-		if ( $header[0] =~m{^SIP/2.0\s+(\d+)\s+(\S.*?)\s*$} ) {
-			# Response, e.g. SIP/2.0 407 Authorization required
-			$result{code} = $1;
-			$result{text} = $2;
-		} elsif ( $header[0] =~m{^(\w+)\s+(\S.*?)\s+SIP/2\.0\s*$} ) {
-			# Request, e.g. INVITE <sip:bla@fasel> SIP/2.0
-			$result{code} = $1;
-			$result{text} = $2;
+		my @v;
+		if ( $nk eq 'www-authenticate'
+			|| $nk eq 'proxy-authenticate'
+			|| $nk eq 'authorization'
+			|| $nk eq 'proxy-authorization' ) {
+			# don't split on ','
+			@v = $v;
 		} else {
-			die "bad request: starts with '$header[0]'";
-		}
-		shift(@header);
-
-		$result{body} = $body;
-
-		my @hdr;
-		my @lines;
-		while (@header) {
-			my ($k,$v) = $header[0] =~m{^([^\s:]+)\s*:\s*(.*)}
-				or die "bad header line $header[0]";
-			my $line = shift(@header);
-			while ( @header && $header[0] =~m{^\s+(.*)} ) {
-				# continuation line
-				$v .= "\n$1";
-				$line .= shift(@header);
-			}
-			my $nk = _normalize_hdrkey($k);
-
-			my $parse = $key2parser{$nk};
-			my @v = $parse ? $parse->($v,$nk) : _hdrkey_parse_comma_seperated($v,$nk);
-			if ( @v>1 ) {
-				for( my $i=0;$i<@v;$i++ ) {
-					push @hdr, Net::SIP::HeaderPair->new( $k,$v[$i],scalar(@lines),$i );
+			# split on komma (but not if quoted)
+			push @v,'';
+			my $quoted = 0;
+			while (1) {
+				if ( $v =~m{\G(.*?)([\\",])}gc ) {
+					if ( $2 eq "\\" ) {
+						$v[-1].=$1.$2.substr( $v,pos($v),1 );
+						pos($v)++;
+					} elsif ( $2 eq '"' ) {
+						$v[-1].=$1.$2;
+						$quoted = !$quoted;
+					} elsif ( $2 eq ',' ) {
+						# next item if not quoted
+						( $v[-1].=$1 ) =~s{\s+$}{}; # strip trailing space
+						push @v,'' if !$quoted;
+						$v =~m{\G\s+}gc; # skip space after ','
+					}
+				} else {
+					# add rest to last from @v
+					$v[-1].= substr($v,pos($v)||0 );
+					last;
 				}
-			} else {
-				push @hdr, Net::SIP::HeaderPair->new( $k,$v[0],scalar(@lines) );
 			}
-			push @lines, [ $line, int(@v) ];
 		}
-		$result{header} = \@hdr;
-		$result{lines}  = \@lines;
-		return \%result;
+		if ( @v>1 ) {
+			for( my $i=0;$i<@v;$i++ ) {
+				push @hdr, Net::SIP::HeaderPair->new( $k,$v[$i],scalar(@lines),$i );
+			}
+		} else {
+			push @hdr, Net::SIP::HeaderPair->new( $k,$v[0],scalar(@lines) );
+		}
+		push @lines, [ $line, int(@v) ];
 	}
+	$self->{header} = \@hdr;
+	$self->{lines}  = \@lines;
+
+	return @{$self}{qw( code text header body )};
 }
 
 ###########################################################################
@@ -640,7 +560,7 @@ sub as_parts {
 sub sdp_body {
 	my Net::SIP::Packet $self = shift;
 	my $ct = $self->get_header( 'content-type' );
-	return if $ct && lc($ct) ne 'application/sdp';
+	return if $ct && $ct ne 'application/sdp';
 	my $body = ($self->as_parts)[3] || return;
 	return Net::SIP::SDP->new( $body );
 }
@@ -681,8 +601,8 @@ sub _update_string {
 # Args: $key
 # Returns: $key_normalized
 ###########################################################################
-sub _normalize_hdrkey {
-	goto &Net::SIP::HeaderPair::_normalize_hdrkey
+sub _normalize_hdrkey { 
+	goto &Net::SIP::HeaderPair::_normalize_hdrkey 
 }
 
 ###########################################################################
@@ -765,7 +685,7 @@ sub set_modified {
 
 ###########################################################################
 # Net::SIP::HeaderVal;
-# gives string representation and hash representation
+# gives string representation and hash representation 
 # (split by ';' or ',') of header value
 ###########################################################################
 
