@@ -45,13 +45,6 @@ sub new {
 	return $self;
 }
 
-# hack to have access to the store, to dump or restore it
-sub _store { 
-	my $self = shift;
-	$self->{store} = shift if @_;
-	return $self->{store};
-}
-
 ###########################################################################
 # handle packet, called from Net::SIP::Dispatcher on incoming requests
 # Args: ($self,$packet,$leg,$addr)
@@ -69,37 +62,22 @@ sub receive {
 
 	# accept only REGISTER
 	$packet->is_request || return;
-	if ( $packet->method ne 'REGISTER' ) {
-		# if we know the target rewrite the destination URI
-		my $addr = (sip_uri2parts($packet->uri))[3];
-		DEBUG( 1,"method ".$packet->method." addr=<$addr>" );
-		my @found = $self->query( $addr );
-		@found or do {
-			DEBUG( 1, "$addr not locally registered" );
-			return;
-		};
-		DEBUG( 1,"rewrite URI $addr in ".$packet->method." to $found[0]" );
-		$packet->set_uri( $found[0] );
-		return; # propagate to next in chain
-	}
+	$packet->method eq 'REGISTER' || return;
 
 	my $from = $packet->get_header( 'from' ) or do {
-		DEBUG( 1,"no from in register request. DROP" );
+		DEBUG( "no from in register" );
 		return;
 	};
 
 	# what address will be registered
 	($from) = sip_hdrval2parts( from => $from );
-	if ( my ($domain,$user,$proto) = sip_uri2parts( $from ) ) {
-		# normalize if possible
-		$from = "$proto:$user\@$domain";
-	}
+	$from = $1 if $from =~m{<(sips?:\S+)>}i;
 
 	# check if domain is allowed
 	if ( my $rd = $self->{domains} ) {
 		my ($domain) = $from =~m{\@([\w\-\.]+)};
 		if ( ! first { $domain =~m{\.?\Q$_\E$}i || $_ eq '*' } @$rd ) {
-			DEBUG( 1, "$domain matches none of my own domains. DROP" );
+			DEBUG( "$domain matches none of my own domains" );
 			return;
 		}
 	}
@@ -111,15 +89,16 @@ sub receive {
 
 	# to which contacs it will be registered
 	my @contact = $packet->get_header( 'contact' );
+	my $store = $self->{store};
+	my $curr = $store->{ $from } ||= {};
 
-	my %curr;
 	foreach my $c (@contact) {
 		# update contact info
 		my ($c_addr,$param) = sip_hdrval2parts( contact => $c );
 		$c_addr = $1 if $c_addr =~m{<(\w+:\S+)>}; # do we really need this?
 		my $expire = $param->{expires};
 		$expire = $glob_expire if ! defined $expire;
-		$expire = $self->{max_expires}
+		$expire = $self->{max_expires} 
 			if ! defined $expire || $expire > $self->{max_expires};
 		if ( $expire ) {
 			if ( $expire < $self->{min_expires} ) {
@@ -132,18 +111,16 @@ sub receive {
 			}
 			$expire += $now if $expire;
 		}
-		$curr{$c_addr} = $expire;
+		$curr->{$c_addr} = $expire;
 	}
-
-	$self->{store}{ $from } = \%curr;
-
+	
 	# expire now!
 	$self->expire();
-	DEBUG_DUMP( 100,$self->{store} );
+	DEBUG_DUMP( $store );
 
 	# send back a list of current contacts
 	my $response = $packet->create_response( '200','OK' );
-	while ( my ($where,$expire) = each %curr ) {
+	while ( my ($where,$expire) = each %$curr ) {
 		$expire -= $now;
 		$response->add_header( contact => "<$where>;expires=$expire" );
 	}
@@ -153,19 +130,6 @@ sub receive {
 	return 200;
 }
 
-###########################################################################
-# return information for SIP address
-# Args: ($self,$addr)
-# Returns: @sip_contacts
-###########################################################################
-sub query {
-	my Net::SIP::Registrar $self = shift;
-	my $addr = shift;
-	DEBUG( 50,"lookup of $addr" );
-	my $contacts = $self->{store}{$addr} || return;
-	return grep { m{^sips?:} } keys %$contacts;
-}
-
 ###########################################################################
 # remove all expired entries from store
 # Args: $self
