@@ -10,6 +10,7 @@ use warnings;
 
 package Net::SIP::Simple::Call;
 use base 'Net::SIP::Simple';
+
 use fields qw( call_cleanup rtp_cleanup ctx param );
 
 ###########################################################################
@@ -35,38 +36,22 @@ use fields qw( call_cleanup rtp_cleanup ctx param );
 #   send_bye: callback or scalar-ref used when call is closed by local side
 #   sdp_peer: Net::SIP::SDP from peer
 #   clear_sdp: ' causes that keys sdp,sdp_peer,media_ssocks and
-#       media_lsocks gets cleared on new invite, so that a new SDP session
+#       media_lsocks gets cleared on new invite, so that a new SDP session 
 #       need to be established
 #   cb_final: callback which will be called on final response in INVITE
-#       with (status,self,%args) where status is OK|FAIL
-#   cb_preliminary: callback which will be called on preliminary response
-#       in INVITE with (self,code,packet)
+#       with (result,status,self,%args) where status is OK|FAIL
 #   cb_established: callback which will be called on receiving ACK in INVITE
-#       with (status,self) where status is OK|FAIL
-#   cb_invite: callback called with ($self,$packet) when INVITE is received
-#   sip_header: hashref of SIP headers to add
-#   call_on_hold: one-shot parameter to set local media addr to 0.0.0.0,
-#       will be set to false after use
-#   rtp_param: [ pt,size,interval,name ] RTP payload type, packet size and interval
-#       between packets managed in Net::SIP::Simple::RTP, default is PCMU/8000,
-#       e.g [ 0,160,160/8000 ]
-#       a name can be added in which case an rtpmap and ptme entry will be created in the
-#       SDP, e.g. [ 97,240,0.03,'iLBC/8000' ]
-###########################################################################
+#       with (result,status,self) where status is OK|FAIL
 
 use Net::SIP::Util qw(create_rtp_sockets invoke_callback);
 use Net::SIP::Debug;
 use Socket;
-use Storable 'dclone';
-use Carp 'croak';
-use Scalar::Util 'weaken';
 
 ###########################################################################
 # create a new call based on a controller
 # Args: ($class,$control,$ctx;$param)
 #   $control: Net::SIP::Simple object which controls this call
 #   $ctx: SIP address of peer for new call or NET::SIP::Endpoint::Context
-#        or hashref for constructing NET::SIP::Endpoint::Context
 #   $param: see description of field 'param'
 # Returns: $self
 ###########################################################################
@@ -74,20 +59,15 @@ sub new {
 	my ($class,$control,$ctx,$param) = @_;
 	my $self = fields::new( $class );
 	%$self = %$control;
-
-	$self->{ua_cleanup} = [];
-	$ctx = { to => $ctx } if ! ref($ctx);
-	$ctx->{from}    ||= $self->{from};
-	$ctx->{contact} ||= $self->{contact};
-	$ctx->{auth}    ||= $self->{auth};
-	$ctx->{route}   ||= $self->{route};
-	$self->{ctx} = $ctx;
-
+	$self->{ctx} = ref($ctx) ? $ctx : {
+		to => $ctx,
+		from => $self->{from},
+		auth => $self->{auth},
+	};
 	$self->{call_cleanup} = [];
 	$self->{rtp_cleanup}  = [];
 	$self->{param} = $param ||= {};
 	$param->{init_media} ||= $self->rtp( 'media_recv_echo' );
-	$param->{rtp_param}  ||= [ 0,160,160/8000 ]; # PCMU/8000: 5*160 bytes/second
 	return $self;
 }
 
@@ -103,11 +83,8 @@ sub cleanup {
 	while ( my $cb = shift @{ $self->{call_cleanup} } ) {
 		invoke_callback($cb,$self)
 	}
-	if ( my $ctx = $self->{ctx} ) {
-		$self->{endpoint}->close_context( $ctx );
-	}
-	$self->{param} = {};
-	$self->SUPER::cleanup;
+	%$self = ();
+	DEBUG( "done" );
 }
 
 sub rtp_cleanup {
@@ -115,47 +92,14 @@ sub rtp_cleanup {
 	while ( my $cb = shift @{ $self->{rtp_cleanup} } ) {
 		invoke_callback($cb,$self)
 	}
-	DEBUG( 100,"done" );
+	DEBUG( "done" );
 }
 
 sub DESTROY {
-	DEBUG( 100,"done" );
+	DEBUG( "done" );
 }
-
-
-###########################################################################
-# return peer of call
-# Args: $self
-# Returns: $peer
-###########################################################################
-sub get_peer {
-	my Net::SIP::Simple::Call $self = shift;
-	return $self->{ctx}->peer;
-}
-
-###########################################################################
-# set parameter
-# Args: ($self,%param)
-# Returns: $self
-###########################################################################
-sub set_param {
-	my Net::SIP::Simple::Call $self = shift;
-	my %args = @_;
-	@{ $self->{param} }{ keys %args } = values %args;
-	return $self;
-}
-
-###########################################################################
-# get value for parameter(s)
-# Args: ($self,@keys)
-# Returns: @values|$value[0]
-###########################################################################
-sub get_param {
-	my Net::SIP::Simple::Call $self = shift;
-	my @v = @{$self->{param}}{@_};
-	return wantarray ? @v : $v[0];
-}
-
+		
+	
 ###########################################################################
 # (Re-)Invite other party
 # Args: ($self;%param)
@@ -175,7 +119,7 @@ sub reinvite {
 	$clear_sdp = $param->{clear_sdp} if ! defined $clear_sdp;
 	if ( $clear_sdp ) {
 		# clear SDP keys so that a new SDP session will be created
-		@{ $param }{qw( sdp _sdp_saved sdp_peer media_ssocks media_lsocks )} = ()
+		@{ $param }{qw( sdp sdp_peer media_ssocks media_lsocks )} = ()
 	}
 	$self->{param} = $param = { %$param, %args } if %args;
 
@@ -196,144 +140,55 @@ sub reinvite {
 
 	# predefined callback
 	my $cb = sub {
-		my Net::SIP::Simple::Call $self = shift || return;
-		my ($endpoint,$ctx,$errno,$code,$packet,$leg,$from,$ack) = @_;
-
+		my ($self,$param,$endpoint,$ctx,$errno,$code,$packet,$leg,$from,$ack) = @_;
 		if ( $errno ) {
-			if ( $code == 487 ) {
-				# request was canceled, probably be me -> ignore
+			$self->error( "Failed with error $errno" );
+			invoke_callback( $param->{cb_final}, 'FAIL',$self,errno => $errno );
+			return;
+		}
+		if ( $packet->is_response ) {
+			# response to INVITE
+			# all other responses will not be propagated to this callback
+			if ( $code =~m{^2\d\d} ) {
+
+				# cleanup RTP from last call
+				$self->rtp_cleanup;
+
+				$self->_setup_peer_rtp_socks( $packet ) || do {
+					invoke_callback( $param->{cb_final},'FAIL',$self );
+					return;
+				};
+				if ( $param->{sdp_on_ack} && $ack ) {
+					$self->_setup_local_rtp_socks;
+					$ack->set_body( $param->{sdp} );
+				}
+				invoke_callback( $param->{cb_final},'OK',$self );
+				invoke_callback( $param->{init_media},$self,$param );
+
 			} else {
-				$self->error( "Failed with error $errno".( $code ? " code=$code" :"" ) );
+				invoke_callback( $param->{cb_final},'FAIL',$self,code => $code );
+			}
+
+		} else {
+			if ( $packet->method eq 'BYE' ) {
+				# Hangup
+				invoke_callback( $param->{recv_bye},$param);
+				$endpoint->close_context($packet);
 			}
-			invoke_callback( $param->{cb_final}, 'FAIL',$self,errno => $errno,
-			    code => $code,packet => $packet );
-			return;
 		}
-
-		# new requests in existing call are handled in receive()
-		return $self->receive( @_ ) if $packet->is_request;
-
-		# response to INVITE
-		# all other responses will not be propagated to this callback
-		my $param = $self->{param};
-		if ( $code =~m{^1\d\d} ) {
-			# preliminary response, ignore
-			DEBUG(10,"got preliminary response of %s|%s to INVITE",$code,$packet->msg );
-			invoke_callback( $param->{cb_preliminary},$self,$code,$packet );
-			return;
-		} elsif ( $code !~m{^2\d\d} ) {
-			DEBUG(10,"got response of %s|%s to INVITE",$code,$packet->msg );
-			invoke_callback( $param->{cb_final},'FAIL',$self,code => $code,
-			    packet => $packet );
-			return;
-		}
-
-		# cleanup RTP from last call
-		$self->rtp_cleanup;
-
-		$self->_setup_peer_rtp_socks( $packet ) || do {
-			invoke_callback( $param->{cb_final},'FAIL',$self );
-			return;
-		};
-		if ( $param->{sdp_on_ack} && $ack ) {
-			$self->_setup_local_rtp_socks;
-			$ack->set_body( $param->{sdp} );
-		}
-		invoke_callback( $param->{cb_final},'OK',$self, packet => $packet );
-		invoke_callback( $param->{init_media},$self,$param );
 	};
 
-
 	my $stopvar = 0;
 	$param->{cb_final} ||= \$stopvar;
-	$cb = [ $cb,$self ];
-	weaken( $cb->[1] );
-	$self->{ctx} = $self->{endpoint}->invite(
-		$ctx, $cb, $sdp,
-		$param->{sip_header} ? %{ $param->{sip_header} } : ()
-	);
-
+	$self->{ctx} = $self->{endpoint}->invite( $ctx, [ $cb,$self,$param ], $sdp );
 	if ( $param->{cb_final} == \$stopvar ) {
-
-		# This callback will be called on timeout or response to cancel which
-		# got send after ring_time was over
-		my $noanswercb;
-		if ( $param->{ring_time} ) {
-			$noanswercb = sub {
-				my Net::SIP::Simple::Call $self = shift || return;
-				my ($endpoint,$ctx,$errno,$code,$packet,$leg,$from,$ack) = @_;
-
-				$stopvar = 'NOANSWER' ;
-				my $param = $self->{param};
-				invoke_callback( $param->{cb_noanswer}, 'NOANSWER',$self,
-					errno => $errno,code => $code,packet => $packet );
-
-				if ( $code =~ m{^2\d\d} ) {
-					DEBUG(10,"got response of %s|%s to CANCEL",$code,$packet->msg );
-					invoke_callback( $param->{cb_final},'NOANSWER',$self,code => $code, 
-						packet => $packet );
-				}
-			};
-			$noanswercb = [ $noanswercb,$self ];
-			weaken( $noanswercb->[1] );
-
-			# wait until final response
-			$self->loop( $param->{ring_time}, \$stopvar );
-
-			unless ($stopvar) { # timed out
-				$self->{endpoint}->cancel_invite( $self->{ctx},undef, $noanswercb );
-				$self->loop( \$stopvar );
-			}
-		} else {
-			# wait until final response
-			$self->loop( \$stopvar );
-		}
-
-		$param->{cb_final} = undef;
+		# wait until final response
+		$self->loop( \$stopvar );
 	}
 	return $self->{ctx};
 }
 
 
-###########################################################################
-# cancel call
-# Args: ($self,%args)
-#   %args:
-#     cb_final: callback when CANCEL was delivered. If not given send_cancel
-#        callback on Call object will be used
-# Returns: true if call could be canceled
-# Comment: cb_final gets triggered if the reply for the CANCEL is received
-# or waiting for the reply timed out
-###########################################################################
-sub cancel {
-	my Net::SIP::Simple::Call $self = shift;
-	my %args = @_;
-
-	my $cb = delete $args{cb_final};
-	%args = ( %{ $self->{param} }, %args );
-	$cb ||= $args{send_cancel};
-
-	my $cancel_cb = [
-		sub {
-			my Net::SIP::Simple::Call $self = shift || return;
-			my ($cb,$args,$endpoint,$ctx,$error,$code) = @_;
-			# we don't care about the cause of this callback
-			# it might be a successful or failed reply packet or no reply
-			# packet at all (timeout) - the call is considered closed
-			# in any case except for 1xx responses
-			if ( $code && $code =~m{^1\d\d} ) {
-				DEBUG( 10,"got prelimary response for CANCEL" );
-				return;
-			}
-			invoke_callback( $cb,$args );
-		},
-		$self,$cb,\%args
-	];
-	weaken( $cancel_cb->[1] );
-
-	return $self->{endpoint}->cancel_invite( $self->{ctx}, undef, $cancel_cb );
-}
-
 ###########################################################################
 # end call
 # Args: ($self,%args)
@@ -352,27 +207,27 @@ sub bye {
 	%args = ( %{ $self->{param} }, %args );
 	$cb ||= $args{send_bye};
 
-	my $bye_cb = [
+	my $bye_cb = [ 
 		sub {
-			my Net::SIP::Simple::Call $self = shift || return;
-			my ($cb,$args,$endpoint,$ctx,$error,$code) = @_;
+			my ($self,$cb,$args,$endpoint,$ctx,$error,$code) = @_;
+			DEBUG( "got response for BYE: $code" );
 			# we don't care about the cause of this callback
 			# it might be a successful or failed reply packet or no reply
 			# packet at all (timeout) - the call is considered closed
 			# in any case except for 1xx responses
 			# FIXME: should we check for 302 moved etc?
 			if ( $code && $code =~m{^1\d\d} ) {
-				DEBUG( 10,"got prelimary response for BYE" );
+				DEBUG( "got prelimary response for BYE" );
 				return;
 			}
 			invoke_callback( $cb,$args );
 			$self->cleanup;
-		},
-		$self,$cb,\%args
+		}, 
+		$self,$cb,\%args 
 	];
-	weaken( $bye_cb->[1] );
-
+		
 	$self->{endpoint}->new_request( 'BYE',$self->{ctx}, $bye_cb );
+	DEBUG( "send BYE" );
 }
 
 ###########################################################################
@@ -404,49 +259,38 @@ sub receive {
 		} elsif ( $method eq 'ACK' || $method eq 'INVITE' ) {
 
 			# can transport sdp data
-			if ( my $sdp_peer = eval { $packet->sdp_body } ) {
-				DEBUG( 50,"got sdp data from peer: ".$sdp_peer->as_string );
+			if ( my $sdp_peer = $packet->sdp_body ) {
+				DEBUG( "got sdp data from peer: ".$sdp_peer->as_string );
 				$self->_setup_peer_rtp_socks( $sdp_peer );
-			} elsif ($@) {
-				# mailformed SDP?
-				DEBUG(10,"SDP parsing failed, ignoring packet: $@");
-				return;
 			}
 
 			if ( $method eq 'INVITE' ) {
 
 				if ( $param->{clear_sdp} ) {
 					# clear SDP keys so that a new SDP session will be created
-					@{ $param }{qw( sdp _sdp_saved sdp_peer media_ssocks media_lsocks )} = ()
+					@{ $param }{qw( sdp sdp_peer media_ssocks media_lsocks )} = ()
 				}
 
 				$param->{leg} ||= $leg;
 				$self->_setup_local_rtp_socks;
-				invoke_callback($param->{cb_invite},$self,$packet);
 
 				# send 200 OK with sdp body
 				my $response = $packet->create_response(
 					'200','OK',{},$param->{sdp} );
-				DEBUG( 100,'created response '.$response->as_string );
+				DEBUG( 'created response '.$response->as_string );
 				$self->{endpoint}->new_response( $ctx,$response,$leg,$from );
 
 			} elsif ( $method eq 'ACK' ) {
-				$self->rtp_cleanup; # close last RTP session
 				invoke_callback($param->{cb_established},'OK',$self);
 				invoke_callback($param->{init_media},$self,$param);
 			}
-
-		} elsif ( $method eq 'OPTIONS' ) {
-
-			my $response = $packet->create_response( '200','OK',$self->{options} );
-			$self->{endpoint}->new_response( $ctx,$response,$leg,$from );
 		}
 
 	} else {
 		# don't expect any responses.
 		# Response to BYE is handled by Net::SIP::Endpoint::Context
 		# other responses from the peer I don't expect
-		DEBUG( 100,"got response. WHY? DROP." );
+		DEBUG( "got response. WHY? DROP." );
 	}
 }
 
@@ -482,18 +326,12 @@ sub _setup_peer_rtp_socks {
 	}
 
 	my $raddr = $param->{media_raddr} = [];
-	my $null_address = pack( 'CCCC',0,0,0,0 ); # c=0.0.0.0 => call on hold
 	foreach my $m (@media) {
 		my $range = $m->{range} || 1;
-		my $paddr = inet_aton( $m->{addr} );
-		if ( $paddr eq $null_address ) {
-			# on-hold for this media
-			push @$raddr, undef;
-		} else {
-			my @socks = map { scalar(sockaddr_in( $m->{port}+$_ , $paddr )) }
-				(0..$range-1);
-			push @$raddr, @socks == 1 ? $socks[0] : \@socks;
-		}
+		my @socks = map {
+			scalar(sockaddr_in( $m->{port} + $_ ,inet_aton( $m->{addr} )));
+		} (0..$range-1);
+		push @$raddr, @socks == 1 ? $socks[0] : \@socks;
 	}
 
 	return 1;
@@ -509,10 +347,7 @@ sub _setup_local_rtp_socks {
 	my Net::SIP::Simple::Call $self = shift;
 	my $param = $self->{param};
 
-	my $call_on_hold = $param->{call_on_hold};
-	$param->{call_on_hold} = 0; # one-shot
-
-	my $sdp = $param->{_sdp_saved} || $param->{sdp};
+	my $sdp = $param->{sdp};
 	if ( $sdp && !UNIVERSAL::isa( $sdp,'Net::SIP::SDP' )) {
 		$sdp = Net::SIP::SDP->new( $sdp );
 	}
@@ -522,7 +357,7 @@ sub _setup_local_rtp_socks {
 		# create SDP body
 		my $raddr = $param->{media_rsocks};
 
-		# if no raddr yet just assume one
+		# if no raddr yet just assume one 
 		my @media;
 		if ( my $sdp_peer = $param->{sdp_peer} ) {
 			foreach my $m ( $sdp_peer->get_media ) {
@@ -530,10 +365,6 @@ sub _setup_local_rtp_socks {
 					$self->error( "only RTP/AVP supported" );
 					return;
 				}
-				if ( $m->{media} eq 'audio' ) {
-					# enforce PCMU/8000 for now
-					$m = { %$m, fmt => '0' }
-				}
 				push @media, {
 					media => $m->{media},
 					proto => $m->{proto},
@@ -542,14 +373,10 @@ sub _setup_local_rtp_socks {
 				};
 			}
 		} else {
-			my $rp = $param->{rtp_param};
 			push @media, {
 				proto => 'RTP/AVP',
 				media => 'audio',
-				fmt   => $rp->[0] || 0, # PCMU/8000
-				$rp->[3] ? (
-					a => [ "rtpmap:$rp->[0] $rp->[3]" , "ptime:".$rp->[2]*1000 ]
-				) :(),
+				fmt   => 0, # PCMU/8000
 			}
 		}
 
@@ -561,7 +388,7 @@ sub _setup_local_rtp_socks {
 			$m->{port} = $port;
 		}
 
-		$sdp = $param->{sdp} = Net::SIP::SDP->new(
+		$sdp = $param->{sdp} = Net::SIP::SDP->new( 
 			{ addr => $laddr },
 			@media
 		);
@@ -575,8 +402,8 @@ sub _setup_local_rtp_socks {
 	# asymetric_rtp, e.g. source socket of packet to peer is not the socket where RTP
 	# from peer gets received
 	if ( !$param->{media_ssocks} && $param->{asymetric_rtp} ) {
-		my @arg = (
-			Proto => 'udp',
+		my @arg = ( 
+			Proto => 'udp', 
 			LocalAddr => ( $param->{rtp_addr} || $laddr )
 		);
 		my $msocks = $param->{media_ssocks} = [];
@@ -593,14 +420,6 @@ sub _setup_local_rtp_socks {
 			push @$msocks,$socks;
 		}
 	}
-
-	$param->{_sdp_saved} = $sdp;
-	if ( $call_on_hold ) {
-		$sdp = dclone($sdp); # make changes on clone
-		my @new = map { [ '0.0.0.0',$_->{port} ] } $sdp->get_media;
-		$sdp->replace_media_listen( @new );
-		$param->{sdp} = $sdp;
-	}
 }
 
 1;
